import os
import posixpath
import tempfile
import time
import typing
import uuid

import pandas as pd
from mlflow.entities import RunLog
from mlflow.tracking import MlflowClient

from mlfoundry import log_types
from mlfoundry.constants import RUN_LOGS_DIR
from mlfoundry.dataset.serde_utils import (
    load_dataframe,
    load_series,
    save_dataframe,
    save_series,
)
from mlfoundry.dataset.types import DatasetSchema, DatasetStats, Profiles
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.logger import logger

DATASET_LOG_DIR = posixpath.join(RUN_LOGS_DIR, "dataset")


class TabularDatasetSerDe:
    _FEATURE_ENTITY_NAME = "features"
    _PREDICTIONS_ENTITY_NAME = "predictions"
    _ACTUALS_ENTITY_NAME = "actuals"
    _SCHEMA_FILE_NAME = "schema.json"
    _STATS_FILE_NAME = "stats.json"
    _FEATURES_PROFILE_FILE_NAME = "features_profile.pb.bin"
    _PREDICTIONS_PROFILE_FILE_NAME = "predictions_profile.pb.bin"
    _ACTUALS_PROFILE_FILE_NAME = "actuals_profile.pb.bin"

    def __init__(self, mlflow_client: MlflowClient):
        self.mlflow_client: MlflowClient = mlflow_client
        self.dataset_prefix = DATASET_LOG_DIR

    def _load_file(
        self,
        run_id: str,
        artifact_path: str,
        deserializer: typing.Callable,
    ):
        with tempfile.TemporaryDirectory() as local_dir:
            local_path = self.mlflow_client.download_artifacts(
                run_id, artifact_path, local_dir
            )
            return deserializer(local_path)

    def _get_run_log(
        self, run_id: str, dataset_name: str, log_type: str
    ) -> typing.Optional[RunLog]:
        run_logs = self.mlflow_client.list_run_logs(
            run_uuid=run_id, key=dataset_name, log_type=log_type
        )
        if len(run_logs) == 0:
            return None
        return run_logs[0]

    def validate_dataset_is_not_already_logged(self, run_id, dataset_name):
        run_log = self._get_run_log(
            run_id=run_id,
            dataset_name=dataset_name,
            log_type=log_types.DatasetArtifact.get_log_type(),
        )
        if run_log is None:
            return
        raise MlFoundryException(f"dataset {dataset_name} is already logged")

    def save_dataset(
        self,
        run_id: str,
        dataset_name: str,
        features: pd.DataFrame,
        predictions: typing.Optional[pd.Series],
        actuals: typing.Optional[pd.Series],
        schema: DatasetSchema,
        stats: typing.Optional[DatasetStats],
        profiles: typing.Optional[Profiles],
        only_stats: bool,
    ):
        artifact_path = posixpath.join(
            self.dataset_prefix, dataset_name, str(uuid.uuid4())
        )
        timestamp = int(time.time() * 1000)
        run_logs = [
            log_types.DatasetArtifact(artifact_path=artifact_path).to_run_log(
                key=dataset_name, timestamp=timestamp
            )
        ]
        with tempfile.TemporaryDirectory() as local_dir:
            if not only_stats:
                file_name, file_format = save_dataframe(
                    local_dir=local_dir,
                    entity_name=TabularDatasetSerDe._FEATURE_ENTITY_NAME,
                    dataframe=features,
                )
                run_logs.append(
                    log_types.DatasetArtifactFeatures(
                        artifact_path=posixpath.join(artifact_path, file_name),
                        format=file_format,
                    ).to_run_log(key=dataset_name, timestamp=timestamp)
                )

                if predictions is not None:
                    file_name, file_format = save_series(
                        local_dir=local_dir,
                        entity_name=TabularDatasetSerDe._PREDICTIONS_ENTITY_NAME,
                        series=predictions,
                    )
                    run_logs.append(
                        log_types.DatasetArtifactPredictions(
                            artifact_path=posixpath.join(artifact_path, file_name),
                            format=file_format,
                        ).to_run_log(key=dataset_name, timestamp=timestamp)
                    )
                if actuals is not None:
                    file_name, file_format = save_series(
                        local_dir=local_dir,
                        entity_name=TabularDatasetSerDe._ACTUALS_ENTITY_NAME,
                        series=actuals,
                    )
                    run_logs.append(
                        log_types.DatasetArtifactActuals(
                            artifact_path=posixpath.join(artifact_path, file_name),
                            format=file_format,
                        ).to_run_log(key=dataset_name, timestamp=timestamp)
                    )

            run_logs.append(
                schema.to_log().to_run_log_as_artifact(
                    key=dataset_name,
                    run_id=run_id,
                    mlflow_client=self.mlflow_client,
                    file_name=TabularDatasetSerDe._SCHEMA_FILE_NAME,
                    artifact_path=artifact_path,
                )
            )
            if stats:
                run_logs.append(
                    stats.to_log().to_run_log_as_artifact(
                        key=dataset_name,
                        run_id=run_id,
                        mlflow_client=self.mlflow_client,
                        file_name=TabularDatasetSerDe._STATS_FILE_NAME,
                        artifact_path=artifact_path,
                    )
                )
            if profiles:
                features_profile_path = os.path.join(
                    local_dir, TabularDatasetSerDe._FEATURES_PROFILE_FILE_NAME
                )
                profiles.features.write_protobuf(features_profile_path)
                run_logs.append(
                    log_types.DatasetArtifactFeaturesProfile(
                        artifact_path=posixpath.join(
                            artifact_path,
                            TabularDatasetSerDe._FEATURES_PROFILE_FILE_NAME,
                        )
                    ).to_run_log(key=dataset_name, timestamp=timestamp)
                )

                if profiles.predictions is not None:
                    predictions_profile_path = os.path.join(
                        local_dir, TabularDatasetSerDe._PREDICTIONS_PROFILE_FILE_NAME
                    )
                    profiles.predictions.write_protobuf(predictions_profile_path)
                    run_logs.append(
                        log_types.DatasetArtifactPredictionsProfile(
                            artifact_path=posixpath.join(
                                artifact_path,
                                TabularDatasetSerDe._PREDICTIONS_PROFILE_FILE_NAME,
                            )
                        ).to_run_log(key=dataset_name, timestamp=timestamp)
                    )

                if profiles.actuals is not None:
                    actuals_profile_path = os.path.join(
                        local_dir, TabularDatasetSerDe._ACTUALS_PROFILE_FILE_NAME
                    )
                    profiles.actuals.write_protobuf(actuals_profile_path)
                    run_logs.append(
                        log_types.DatasetArtifactActualsProfile(
                            artifact_path=posixpath.join(
                                artifact_path,
                                TabularDatasetSerDe._ACTUALS_PROFILE_FILE_NAME,
                            )
                        ).to_run_log(key=dataset_name, timestamp=timestamp)
                    )

            self.mlflow_client.log_artifacts(
                run_id=run_id, local_dir=local_dir, artifact_path=artifact_path
            )
            self.mlflow_client.insert_run_logs(run_uuid=run_id, run_logs=run_logs)

    def load_schema(self, run_id: str, dataset_name: str) -> DatasetSchema:
        run_log = self.mlflow_client.get_latest_run_log(
            run_uuid=run_id,
            key=dataset_name,
            log_type=log_types.DatasetSchemaRunLog.get_log_type(),
        )
        return DatasetSchema.parse_obj(
            log_types.DatasetSchemaRunLog.from_run_log(run_log=run_log).value
        )

    def load_stats(self, run_id: str, dataset_name: str) -> DatasetStats:
        run_log = self.mlflow_client.get_latest_run_log(
            run_uuid=run_id,
            key=dataset_name,
            log_type=log_types.DatasetStatsRunLog.get_log_type(),
        )
        return DatasetStats.parse_obj(
            log_types.DatasetStatsRunLog.from_run_log(run_log=run_log).value
        )

    def _load_features_profile(self, run_id: str, dataset_name: str):
        from whylogs import DatasetProfile

        run_log = self.mlflow_client.get_latest_run_log(
            run_uuid=run_id,
            key=dataset_name,
            log_type=log_types.DatasetArtifactFeaturesProfile.get_log_type(),
        )
        artifact_path = log_types.DatasetArtifactFeaturesProfile.from_run_log(
            run_log
        ).artifact_path
        return self._load_file(
            run_id=run_id,
            artifact_path=artifact_path,
            deserializer=DatasetProfile.read_protobuf,
        )

    def _load_predictions_profile(self, run_id: str, dataset_name: str):
        from whylogs import DatasetProfile

        run_log = self._get_run_log(
            run_id=run_id,
            dataset_name=dataset_name,
            log_type=log_types.DatasetArtifactPredictionsProfile.get_log_type(),
        )
        if run_log is None:
            logger.info(f"prediction profile for {dataset_name} not found")
            return None

        artifact_path = log_types.DatasetArtifactPredictionsProfile.from_run_log(
            run_log
        ).artifact_path
        return self._load_file(
            run_id=run_id,
            artifact_path=artifact_path,
            deserializer=DatasetProfile.read_protobuf,
        )

    def _load_actuals_profile(self, run_id: str, dataset_name: str):
        from whylogs import DatasetProfile

        run_log = self._get_run_log(
            run_id=run_id,
            dataset_name=dataset_name,
            log_type=log_types.DatasetArtifactActualsProfile.get_log_type(),
        )
        if run_log is None:
            logger.info(f"actuals profile for {dataset_name} not found")
            return None

        artifact_path = log_types.DatasetArtifactActualsProfile.from_run_log(
            run_log
        ).artifact_path
        return self._load_file(
            run_id=run_id,
            artifact_path=artifact_path,
            deserializer=DatasetProfile.read_protobuf,
        )

    def load_profiles(self, run_id: str, dataset_name: str) -> Profiles:
        features_profile = self._load_features_profile(run_id, dataset_name)
        actuals_profile = self._load_actuals_profile(run_id, dataset_name)
        predictions_profile = self._load_predictions_profile(run_id, dataset_name)
        return Profiles(
            features=features_profile,
            actuals=actuals_profile,
            predictions=predictions_profile,
        )

    def load_features(
        self, run_id: str, dataset_name: str
    ) -> typing.Optional[pd.DataFrame]:
        run_log = self._get_run_log(
            run_id=run_id,
            dataset_name=dataset_name,
            log_type=log_types.DatasetArtifactFeatures.get_log_type(),
        )
        if run_log is None:
            logger.info(f"feature of dataset {dataset_name} not found")
            return None

        artifact_path = log_types.DatasetArtifactFeatures.from_run_log(
            run_log
        ).artifact_path
        return self._load_file(
            run_id=run_id, artifact_path=artifact_path, deserializer=load_dataframe
        )

    def load_actuals(
        self, run_id: str, dataset_name: str
    ) -> typing.Optional[pd.DataFrame]:
        run_log = self._get_run_log(
            run_id=run_id,
            dataset_name=dataset_name,
            log_type=log_types.DatasetArtifactActuals.get_log_type(),
        )
        if run_log is None:
            logger.info(f"actuals of dataset {dataset_name} not found")
            return None

        artifact_path = log_types.DatasetArtifactActuals.from_run_log(
            run_log
        ).artifact_path
        return self._load_file(
            run_id=run_id, artifact_path=artifact_path, deserializer=load_series
        )

    def load_predictions(
        self, run_id: str, dataset_name: str
    ) -> typing.Optional[pd.DataFrame]:
        run_log = self._get_run_log(
            run_id=run_id,
            dataset_name=dataset_name,
            log_type=log_types.DatasetArtifactPredictions.get_log_type(),
        )
        if run_log is None:
            logger.info(f"actuals of dataset {dataset_name} not found")
            return None

        artifact_path = log_types.DatasetArtifactPredictions.from_run_log(
            run_log
        ).artifact_path
        return self._load_file(
            run_id=run_id, artifact_path=artifact_path, deserializer=load_series
        )



import copy
import datetime
import json
import os
import tempfile
from pathlib import Path
from typing import Any, Dict, NamedTuple, Optional, Tuple, Union

from mlflow.entities import Artifact, ArtifactType
from mlflow.entities import ArtifactVersion as _ArtifactVersion
from mlflow.tracking import MlflowClient
from pydantic import BaseModel, Extra

from mlfoundry.artifact.truefoundry_artifact_repo import MlFoundryArtifactsRepository
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.artifacts.constants import INTERNAL_METADATA_PATH
from mlfoundry.log_types.artifacts.utils import (
    _get_mlflow_client,
    _validate_artifact_metadata,
    _validate_description,
)
from mlfoundry.logger import logger

# TODO: Add some progress indicators for upload and download
# TODO: Support async download and upload


class ArtifactPath(NamedTuple):
    src: str
    dest: str = None


class ArtifactVersionInternalMetadata(BaseModel):
    class Config:
        extra = Extra.allow

    files_dir: str  # relative to root


class ArtifactVersionDownloadInfo(BaseModel):
    download_dir: str
    content_dir: str


class ArtifactVersion:
    def __init__(self, artifact_version: _ArtifactVersion, artifact: Artifact) -> None:
        self._mlflow_client = _get_mlflow_client()
        self._artifact_version: _ArtifactVersion = artifact_version
        self._artifact: Artifact = artifact
        self._deleted = False
        self._description: str = ""
        self._metadata: Dict[str, Any] = {}
        self._set_mutable_attrs()

    @classmethod
    def from_fqn(cls, fqn: str) -> "ArtifactVersion":
        mlflow_client = _get_mlflow_client()
        artifact_version = mlflow_client.get_artifact_version_by_fqn(fqn=fqn)
        artifact = mlflow_client.get_artifact_by_id(
            artifact_id=artifact_version.artifact_id
        )
        return cls(artifact_version=artifact_version, artifact=artifact)

    def _ensure_not_deleted(self):
        if self._deleted:
            raise MlFoundryException(
                "Artifact Version was deleted, cannot access a deleted version"
            )

    def _set_mutable_attrs(self, refetch=False):
        if refetch:
            self._artifact_version = self._mlflow_client.get_artifact_version_by_id(
                version_id=self._artifact_version.id
            )
        self._description = self._artifact_version.description or ""
        self._metadata = copy.deepcopy(self._artifact_version.artifact_metadata)

    def __repr__(self):
        return f"{self.__class__.__name__}(fqn={self.fqn!r})"

    @property
    def name(self) -> str:
        return self._artifact.name

    @property
    def artifact_fqn(self) -> str:
        return self._artifact.fqn

    @property
    def version(self) -> int:
        return self._artifact_version.version

    @property
    def fqn(self) -> str:
        return self._artifact_version.fqn

    @property
    def step(self) -> int:
        return self._artifact_version.step

    @property
    def description(self) -> Optional[str]:
        return self._description

    @description.setter
    def description(self, value: str):
        _validate_description(value)
        self._description = value

    @property
    def metadata(self) -> Dict[str, Any]:
        return self._metadata

    @metadata.setter
    def metadata(self, value: Dict[str, Any]):
        _validate_artifact_metadata(value)
        self._metadata = value

    @property
    def created_by(self) -> str:
        return self._artifact_version.created_by

    @property
    def created_at(self) -> datetime.datetime:
        return self._artifact_version.created_at

    @property
    def updated_at(self) -> datetime.datetime:
        return self._artifact_version.updated_at

    def raw_download(self, path: Optional[Union[str, Path]]) -> str:
        logger.info(
            "Downloading artifact version contents, this might take a while ..."
        )
        mlfa_repo = MlFoundryArtifactsRepository(
            version_id=self._artifact_version.id, mlflow_client=self._mlflow_client
        )
        return mlfa_repo.download_artifacts(artifact_path="", dst_path=path)

    def _download(
        self, path: Optional[Union[str, Path]]
    ) -> Tuple[ArtifactVersionInternalMetadata, str]:
        self._ensure_not_deleted()
        download_dir = self.raw_download(path=path)
        internal_metadata_path = os.path.join(download_dir, INTERNAL_METADATA_PATH)
        if not os.path.exists(internal_metadata_path):
            raise MlFoundryException(
                f"Artifact version seems to be corrupted or in invalid format due to missing artifact metadata. "
                f"You can still use .raw_download(path='/your/path/here') to download and inspect files."
            )
        with open(internal_metadata_path) as f:
            internal_metadata = ArtifactVersionInternalMetadata.parse_obj(json.load(f))
        download_path = os.path.join(download_dir, internal_metadata.files_dir)
        return internal_metadata, download_path

    def download(self, path: Optional[Union[str, Path]]) -> str:
        _, download_path = self._download(path=path)
        return download_path

    def delete(self) -> bool:
        self._ensure_not_deleted()
        self._mlflow_client.delete_artifact_version(
            version_id=self._artifact_version.id
        )
        self._deleted = True
        return True

    def update(self):
        self._ensure_not_deleted()

        self._artifact_version = self._mlflow_client.update_artifact_version(
            version_id=self._artifact_version.id,
            description=self.description,
            artifact_metadata=self.metadata,
        )
        self._set_mutable_attrs()


def calculate_artifact_size(artifact_dir: tempfile.TemporaryDirectory):
    total_size = 0
    for path, dirs, files in os.walk(artifact_dir.name):
        for f in files:
            file_path = os.path.join(path, f)
            total_size += os.stat(file_path).st_size
    return total_size


def _log_artifact_version_helper(
    run,
    name: str,
    artifact_type: ArtifactType,
    artifact_dir: tempfile.TemporaryDirectory,
    mlflow_client: Optional[MlflowClient] = None,
    ml_repo_id: Optional[str] = None,
    description: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    step: int = 0,
):
    if (run and mlflow_client) or (not run and not mlflow_client):
        raise MlFoundryException("Exactly one of run, mlflow_client should be passed")
    if mlflow_client and not ml_repo_id:
        raise MlFoundryException(
            "If mlflow_client is passed, ml_repo_id must also be passed"
        )
    if run:
        mlflow_client: MlflowClient = run.mlflow_client

    version_id = mlflow_client.create_artifact_version(
        experiment_id=int(run._experiment_id) if run else ml_repo_id,
        artifact_type=artifact_type,
        name=name,
    )
    artifacts_repo = MlFoundryArtifactsRepository(
        version_id=version_id, mlflow_client=mlflow_client
    )
    total_size = calculate_artifact_size(artifact_dir)
    try:
        logger.info(
            "Packaging and uploading files to remote with Artifact Size: %.6f MB",
            total_size / 1000000.0,
        )
        artifacts_repo.log_artifacts(local_dir=artifact_dir.name, artifact_path=None)
    except Exception as e:
        mlflow_client.notify_failure_for_artifact_version(version_id=version_id)
        raise MlFoundryException("Failed to log Artifact") from e
    finally:
        artifact_dir.cleanup()
    artifact_version = mlflow_client.finalize_artifact_version(
        version_id=version_id,
        run_uuid=run.run_id if run else None,
        description=description,
        artifact_metadata=metadata,
        data_path=INTERNAL_METADATA_PATH,
        step=step,
        artifact_size=total_size,
    )

    return ArtifactVersion.from_fqn(fqn=artifact_version.fqn)
